/* -*-	Mode:C++ -*- */
/*
 * Copyright (c) 2007 Regents of the SIGNET lab, University of Padova.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University of Padova (SIGNET lab) nor the 
 *    names of its contributors may be used to endorse or promote products 
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED 
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR 
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF 
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#ifndef VID_MODULE_H
#define VID_MODULE_H

#include <module.h>

#define VID_DROP_REASON_UNKNOWN_TYPE "UKT"
#define VID_DROP_REASON_RECEIVED "RCV"
#define VID_DROP_REASON_OUT_OF_SEQUENCE "OOS"
#define VID_DROP_REASON_WRONG_SADDR "WSA"
#define VID_DROP_REASON_WRONG_SPORT "WSP"

#define HDR_VID(P)      (hdr_vid::access(P))

extern packet_t PT_MVID;

typedef struct hdr_vid {

  double ts;  			// < timestamp pacchetto (istante di generazione) >
  long double video_ts; // < timestamp contenuto nel file di Input >
  
  int sn;     			// < sequence number del pacchetto >
						// (usato per individuare perdite o pacchetti fuori sequenza)  
  int video_id;	   		// < sequence number del pacchetto contenuto nel file di Input >
  						// (viene riportato nel file di output)
  bool rftt_valid; 		// < whether the rftt value is valid (a packet must
		    			//   have been received in the opposite direction) 
		     
  double rftt;    		// < Forward Trip Time seen in the opposite direction
						// (opposite with respect to the direction of this
						// packet). This is used to calculate Round Trip Time 
		
  static int offset_;
  inline static int& offset() { return offset_; }
  inline static struct hdr_vid* access(const Packet* p) {
    return (struct hdr_vid*)p->access(offset_);
  }
  
} hdr_vid;


class VidModule;

/**
 * Ogni VideoPacket conterrà le informazioni per la creazione dei pacchetti
 */
class VideoPacket : public Event
{
  public:		
    long double TimeStamp;	
    int VideoId;		
    int Size;

    VideoPacket(long double timestamp,int id,int size){ 
      TimeStamp = timestamp;
      VideoId = id;
      Size = size; 		
    }
};


/**
 * Handler usato da VidModule per gestire la trasmissione dei pacchetti
 */
class VideoTxHandler : public Handler
{
public:
  VideoTxHandler(VidModule *m) : Handler() { module = m; }
	
protected:
  VidModule* module;
  virtual void handle(Event* e);
};


/**
 * Implementa un trasmettitore/ricevitore di pacchetti "Video"
 * al nodo trasmettitore legge da un file contenente il dump del video da trasmettere
 * al nodo ricevitore salva il dump dei pacchetti ricevuti
 */
class VidModule : public Module
{
  friend class VideoTxHandler;

public:

  VidModule();
  ~VidModule();		
		
  virtual void recv(Packet*);
  virtual void recv(Packet* p, Handler* h);
  virtual int command(int argc, const char*const* argv);
  virtual int crLayCommand(ClMessage* m);

  virtual double GetRTT();
  virtual double GetFTT();
  virtual double GetPER();  
  virtual double GetTHR(); 

  virtual void resetStats(); 


protected:

  static int uidcnt_;  // < simulation-unique packet ID > 
  nsaddr_t dstPort_;
  nsaddr_t dstAddr_;

  /** 
   * inizializza un Packet copiando le informazioni necessarie da Event(<-VideoPacket)
   */	
  virtual void initPkt(Packet* p,Event* e);

  /** 
   * Se è specificato un file di Input schedula tutti gli eventi VideoPacket e
   * inizia la trasmissione
   */
  virtual void start();
		

  /** 
   * Ferma la trasmissione
   */
  virtual void stop();

  virtual void updateRTT(double rtt);
  virtual void updateFTT(double ftt);
  virtual void updateThroughput(int bytes, double dt);

  virtual void incrPktLost(int npkts);
  virtual void incrPktRecv();
  virtual void incrPktOoseq();
  virtual void incrPktInvalid();

  double alpha_;      /**< smoothing factor used for statistics using first order IIR filter */
	       
  int txsn;          /**< sequence number of next new packet to be transmitted */
  int hrsn;          /**< highest received sequence number */

  int pkts_recv;     /**< total number of received packets. Packet out
			              of sequence are not counted here. */
  int pkts_ooseq;    /**< total number of packets received out of
			              sequence  */
  int pkts_lost;     /**< total number of lost packets, including
			              packets received out of sequence   */

  int pkts_invalid;  /**< total number of invalid packets received */

  int pkts_last_reset; /**< for error checking after stats are
			                reset. Set to pkts_lost+pkts_recv each time
    			            resetStats is called. */ 

  double rftt;       /**< Forward Trip Time seen for last received packet */ 

  double srtt;       /**< Smoothed Round Trip Time, calculated as for TCP */

  double sftt;       /**< Smoothed Forward Trip Time, calculated as srtt */

  double lrtime;       /**< Time of last packet reception */

  double sthr;       /**< Smoothed throughput calculation */

  

  VideoTxHandler txhandler;	
  FILE* OutputFile;   /**< File di Output dove salvare il Dump dei pacchetti ricevuti */
  FILE* InputFile;    /**< File di Input che contiene il Dump del video da trasmettere */
 

/* Cumulative statistics */
  double sumrtt;
  int rttsamples;

  double sumftt;
  int fttsamples;

  double sumbytes;
  double sumdt;


		
};

#endif /* VID_MODULE_H */
